(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Xml
imports
  Check_Monad
  "$AFP/Real_Impl/Show_Instances"
  Parser_Monad
begin

datatype xml =
  -- {* node-name, attributes, child-nodes, text *}
  XML string "(string \<times> string) list" "xml list" "string option"

datatype xmldoc =
  -- {* header, body *}
  XMLDOC "string" xml

fun tag :: "xml \<Rightarrow> string" where
  "tag (XML name _ _ _) = name"
hide_const (open) tag

fun children :: "xml \<Rightarrow> xml list" where
  "children (XML _ _ cs _) = cs"
hide_const (open) children

fun num_children :: "xml \<Rightarrow> nat" where
  "num_children (XML _ _ cs _) = length cs"
hide_const (open) num_children

text {*An alternative induction scheme for @{typ xml}.*}
lemma
  fixes P :: "xml \<Rightarrow> bool"
  assumes "\<And>n a cs t. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P (XML n a cs t)"
  shows xml_induct [case_names XML, induct type: xml]: "P xml"
    and "\<And>xml. xml \<in> set xmls \<Longrightarrow> P xml"
  using assms by (induct xml and xmls) auto


subsection {* Printing of XML Nodes and Documents *}

instantiation xml :: "show"
begin

definition
  shows_attr :: "string \<times> string \<Rightarrow> shows"
where
  "shows_attr av = (shows (fst av) +@+ ''='' +#+ shows_string (CHR ''\"'' # snd av @ ''\"''))"

definition
  shows_attrs :: "(string\<times>string)list \<Rightarrow> shows"
where
  "shows_attrs as = shows_map (\<lambda>a. '' '' +#+ shows_attr a) as"

fun
  shows_XML_indent :: "string \<Rightarrow> nat \<Rightarrow> xml \<Rightarrow> shows"
where
  "shows_XML_indent ind i (XML n a c t) =
    (ind +#+ ''<'' +#+ shows n +@+ shows_attrs a +@+
      (if c = [] \<and> t = None then shows_string ''/>''
      else (
        ''>'' +#+
        (if c \<noteq> [] then
          shows_map (shows_XML_indent (ind @ replicate i (CHR '' '')) i) c +@+ shows_string ind
        else id) +@+
        (if t = None then id else shows_string (the t)) +@+
        ''</'' +#+ shows n +@+ shows_string ''>'')))"

definition "shows_prec (d::nat) xml = shows_XML_indent [newline] 2 xml"

lemma shows_attr_assoc:
  "(s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r @ t)"
  unfolding shows_attr_def by (cases av) (auto simp: show_defs o_def)

lemma shows_attr_assoc_elt:
  "\<forall>av\<in>set avs. (s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r @ t)"
  using shows_attr_assoc by best

lemma shows_attrs_assoc:
  "shows_attrs as r @ s = shows_attrs as (r @ s)"
  unfolding shows_attrs_def shows_map_assoc [OF shows_attr_assoc_elt [of as "'' ''"]] by simp

lemma assoc_xml':
  "shows_XML_indent ind i xml r @ s = shows_XML_indent ind i xml (r @ s)"
  by (induct xml arbitrary: ind r s) (auto simp: shows_attrs_assoc)

lemma assoc_xml:
  "shows_prec d (xml::xml) r @ s = shows_prec d xml (r @ s)"
  unfolding shows_prec_xml_def by (rule assoc_xml')

standard_shows_list assoc_xml

end

instantiation xmldoc :: "show"
begin

fun shows_xmldoc where
  "shows_xmldoc (XMLDOC h x) = (h +#+ shows_nl +@+ shows x)"

definition "shows_prec (d::nat) doc = shows_xmldoc doc"

lemma assoc_xmldoc:
  "shows_prec d (x::xmldoc) r @ s = shows_prec d x (r @ s)"
by (cases x) (auto simp: shows_prec_xmldoc_def)

standard_shows_list assoc_xmldoc

end

text {*Resolve problem with recursive dependencies in generated code.*}
lemma [code]:
  "HOL.eq (XML name1 attrs1 children1 text1) (XML name2 attrs2 children2 text2) \<longleftrightarrow>
    HOL.eq name1 name2 \<and> HOL.eq attrs1 attrs2 \<and>
    list_all2 HOL.eq children1 children2 \<and> HOL.eq text1 text2"
  by (simp add: list_all2_eq)


subsection {* XML-Parsing *}

definition parse_text :: "string option parser" where
  "parse_text = do {
    ts \<leftarrow> many (op \<noteq> (CHR ''<''));
    let text = trim ts;
    if text = [] then return None
    else return (Some (List.rev (trim (List.rev text))))
  }"

lemma is_parser_parse_text [intro]:
  "is_parser parse_text"
  by (auto simp: parse_text_def)

definition parse_attribute_value :: "string parser" where
  "parse_attribute_value = do {
    exactly [CHR ''\"''];
    v \<leftarrow> many (op \<noteq> CHR ''\"'');
    exactly [CHR ''\"''];
    return v
  }"

lemma is_parser_parse_attribute_value [intro]:
  "is_parser parse_attribute_value"
  by (auto simp: parse_attribute_value_def)

(*Here we fix what is considered a "letter" for tag-names.*)
definition letters :: "char list" where
  "letters =
    ''abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789&;:-''"

definition many_letters :: "string parser" where
  [simp]: "many_letters = manyof letters"

definition parse_name :: "string parser" where
  "parse_name s \<equiv> (do {
    n \<leftarrow> many_letters;
    spaces;
    if n = [] then error (''expected letter '' @ letters @ '' but first symbol is \"'' @ take 1 s @ ''\"'')
    else return n
  }) s"

definition letters_impl :: "char \<Rightarrow> bool" where
  "letters_impl c \<longleftrightarrow>
    c \<ge> CHR ''a'' \<and> c \<le> CHR ''z'' \<or>
    c \<ge> CHR ''A'' \<and> c \<le> CHR ''Z'' \<or>
    c \<ge> CHR ''0'' \<and> c \<le> CHR ''9'' \<or>
    c \<in> set ''_&;:-''"

lemma letters_impl [code]:
  "c \<in> set letters \<longleftrightarrow> letters_impl c"
proof (cases c)
  case (Char n1 n2)
  show ?thesis
    by (simp add: Char letters_impl_def letters_def less_eq_char_def nat_of_char_def, cases n1)
       (cases n2, simp_all)+
qed

lemma many_letters [code_unfold]:
  "many_letters = many letters_impl"
  by (simp add: letters_impl_def [abs_def] manyof_def letters_impl)

lemma is_parser_parse_name [intro]:
  "is_parser parse_name"
proof 
  fix s r x
  assume res: "parse_name s = Inr (x,r)"
  let ?exp = "do {
    n \<leftarrow> many_letters;
    spaces;
    if n = [] then error (''expected letter '' @ letters @ '' but first symbol is \"'' @ take 1 s @ ''\"'')
    else return n
  }"
  have isp: "is_parser ?exp" by auto
  have id: "parse_name s = ?exp s" by (simp add: parse_name_def)
  from isp[unfolded is_parser_def, rule_format, OF res[unfolded id]] 
  show "length r \<le> length s" .
qed

function (sequential)
  parse_attributes :: "(string \<times> string) list parser"
where
  "parse_attributes [] = Error_Monad.return ([], [])" |
  "parse_attributes (c # s) =
    (if c \<in> set ''/>'' then Error_Monad.return ([], c # s)
    else (do {
      k \<leftarrow> parse_name;
      exactly ''='';
      v \<leftarrow> parse_attribute_value;
      atts \<leftarrow> parse_attributes;
      return ((k, v) # atts)
    }) (c # s))"
  by (pat_completeness) auto

termination parse_attributes
proof
  show "wf (measure length)" by simp
next
  fix c s y ts ya tsa yb tsb
  assume pn: "parse_name (c # s) = Inr (y, ts)"
    and oo: "exactly ''='' ts = Inr (ya, tsa)"
    and pav: "parse_attribute_value tsa = Inr (yb, tsb)"  
  have cp: "is_cparser (exactly ''='')" by auto
  from cp [unfolded is_cparser_def] oo have 1: "length ts > length tsa" by auto
  from is_parser_parse_name [unfolded is_parser_def] pn
    have 2: "length (c # s) \<ge> length ts" by force
  from is_parser_parse_attribute_value [unfolded is_parser_def] pav
    have 3: "length tsa \<ge> length tsb" by force
  from 1 2 3
    show "(tsb, c # s) \<in> measure length" 
    by auto
qed

lemma is_parser_parse_attributes [intro]:
  "is_parser parse_attributes"
proof
  fix s r x 
  assume "parse_attributes s = Inr (x, r)"
  thus "length r \<le> length s"
  proof (induct arbitrary: x rule: parse_attributes.induct)
    case (2 c s)
    show ?case
    proof (cases "c \<in> set ''/>''")
      case True
      with 2(2) show ?thesis by simp
    next
      case False
      from False 2(2) obtain y1 s1 where pn: "parse_name (c # s) = Inr (y1, s1)" by (cases "parse_name (c # s)", auto simp: bind_def)
      from False 2(2) pn obtain y2 s2 where oo: "exactly ''='' s1 = Inr (y2, s2)" by (cases "exactly ''='' s1", auto simp: bind_def)
      from False 2(2) pn oo obtain y3 s3
        where pav: "parse_attribute_value s2 = Inr (y3, s3)"
        by (cases "parse_attribute_value s2") (auto simp: bind_def)
      from False 2(2) pn oo pav obtain y4 s4
        where patts: "parse_attributes s3 = Inr (y4, r)"
        by (cases "parse_attributes s3") (auto simp: return_def bind_def)
      have "length r \<le> length s3" using 2(1)[OF False pn oo pav patts] .
      also have "\<dots> \<le> length s2" using is_parser_parse_attribute_value[unfolded is_parser_def] pav by auto
      also have "\<dots> \<le> length s1" using is_parser_exactly [unfolded is_parser_def] oo by auto
      also have "\<dots> \<le> length (c # s)" using is_parser_parse_name [unfolded is_parser_def] pn by force
      finally show "length r \<le> length (c # s)" by auto
    qed
  qed simp
qed

function parse_nodes :: "xml list parser" where
  "parse_nodes ts = 
    (if ts = [] \<or> take 2 ts = ''</'' then return [] ts
    else (do {
      exactly ''<'';
      n \<leftarrow> parse_name;
      atts \<leftarrow> parse_attributes;
      e \<leftarrow> oneof [''/>'', ''>''];
      (\<lambda> ts'.
        if (e = ''/>'') then (do {
          cs \<leftarrow> parse_nodes;
          return (XML n atts [] None # cs)
        }) ts' else (do {
          t \<leftarrow> parse_text;
          cs \<leftarrow> parse_nodes;
          exactly ''</'';
          exactly n;
          exactly ''>'';
          ns \<leftarrow> parse_nodes;
          return (XML n atts cs t # ns)
        }) ts')
    }) ts)"
  by (pat_completeness) auto

lemma parse_nodes_help:
  "parse_nodes_dom s \<and> (\<forall> x r. parse_nodes s = Inr (x, r) \<longrightarrow> length r \<le> length s)" (is "?prop s")
proof (induct rule: wf_induct [where P = ?prop and r = "measure length"])
  fix s
  assume "\<forall> t. (t, s) \<in> measure length \<longrightarrow> ?prop t"
  hence ind1: "\<And> t. length t < length s \<Longrightarrow> parse_nodes_dom t"  
    and ind2: "\<And> t x r. length t < length s \<Longrightarrow> parse_nodes t = Inr (x,r) \<Longrightarrow> length r \<le> length t" by auto
  let ?check = "\<lambda> s. s = [] \<or> take 2 s = ''</''"
  have dom: "parse_nodes_dom s"
  proof
    fix y 
    assume "parse_nodes_rel y s"
    thus "parse_nodes_dom y"
    proof
      fix ts ya tsa yaa tsb yb tsc yc tsd 
      assume "y = tsd" and "s = ts" and "\<not> ?check ts"
        and "exactly ''<'' ts = Inr (ya, tsa)"
        and "parse_name tsa = Inr (yaa, tsb)"
        and "parse_attributes tsb = Inr (yb, tsc)"
        and "oneof [''/>'', ''>''] tsc = Inr (yc, tsd)"
        and "yc = ''/>''"
      then have len: "length s > length y"
        using is_cparser_exactly [of "''<''"]
        and is_parser_oneof [of "[''/>'', ''>'']"]
        and is_parser_parse_attributes
        and is_parser_parse_name
        by (auto dest!: is_parser_length is_cparser_length)
      with ind1[OF len] show "parse_nodes_dom y" by simp
    next
      fix ts ya tsa yaa tsb yb tsc yc tsd yd tse
      assume "y = tse" and "s = ts" and "\<not> ?check ts"
        and "exactly ''<'' ts = Inr (ya, tsa)"
        and "parse_name tsa = Inr (yaa, tsb)"
        and "parse_attributes tsb = Inr (yb, tsc)"
        and "oneof [''/>'', ''>''] tsc = Inr (yc, tsd)"
        and "parse_text tsd = Inr (yd, tse)"
      then have len: "length s > length y"
        using is_cparser_exactly [of "''<''", simplified]
        and is_parser_oneof [of "[''/>'', ''>'']"]
        and is_parser_parse_attributes
        and is_parser_parse_name
        and is_parser_parse_text
        by (auto dest!: is_parser_length is_cparser_length)
      with ind1[OF len] show "parse_nodes_dom y" by simp
    next
      fix ts ya tsa yaa tsb yb tsc yc tsd yd tse ye tsf yf tsg yg tsh yh tsi yi tsj
      assume y: "y = tsj" and "s = ts" and "\<not> ?check ts"
        and "exactly ''<'' ts = Inr (ya, tsa)"
        and "parse_name tsa = Inr (yaa, tsb)"
        and "parse_attributes tsb = Inr (yb, tsc)"
        and "oneof [''/>'', ''>''] tsc = Inr (yc, tsd)"
        and "parse_text tsd = Inr (yd, tse)"
        and rec: "parse_nodes_sumC tse = Inr (ye, tsf)" 
        and last: "exactly ''</'' tsf = Inr (yf, tsg)"
          "exactly yaa tsg = Inr (yg, tsh)"
          "exactly ''>'' tsh = Inr (yi, tsj)"
      then have len: "length s > length tse"
        using is_cparser_exactly [of "''<''", simplified]
        and is_parser_oneof [of "[''/>'', ''>'']"]
        and is_parser_parse_attributes
        and is_parser_parse_name
        and is_parser_parse_text
        by (auto dest!: is_parser_length is_cparser_length)
      from last(1) last(2) have len2a: "length tsf \<ge> length tsh"
        using is_parser_exactly [of "''</''"] and is_parser_exactly [of yaa]
        and is_parser_parse_name by (auto dest!: is_parser_length)
      have len2c: "length tsh \<ge> length y" using last(3) 
        using is_parser_exactly [of "''>''"] by (auto simp: y dest!: is_parser_length)
      from len2a len2c have len2: "length tsf \<ge> length y" by simp
      from ind2[OF len rec[unfolded parse_nodes_def[symmetric]]] len len2 have "length s > length y" by simp
      from ind1[OF this]       
      show "parse_nodes_dom y" .
    qed
  qed
  note psimps = parse_nodes.psimps[OF dom]
  show "?prop s"
  proof (intro conjI, rule dom, intro allI impI)
    fix x r    
    assume res: "parse_nodes s = Inr (x,r)"
    note res = res[unfolded psimps]
    thus "length r \<le> length s"
    proof (cases "?check s")
      case True      
      thus ?thesis using res by (simp add: return_def)
    next
      case False note oFalse = this
      note res = res[simplified False, simplified]
      from res obtain y1 s1 where oo: "exactly ''<'' s = Inr (y1, s1)" by (cases "exactly ''<'' s", auto simp: bind_def)
      note res = res[unfolded bind_def oo, simplified]
      from res obtain y2 s2 where pn: "parse_name s1 = Inr (y2, s2)" by (cases "parse_name s1", auto simp: bind_def psimps)
      note res = res[simplified bind_def pn, simplified]
      from res obtain y3 s3 where pa: "parse_attributes s2 = Inr (y3, s3)" by (cases "parse_attributes s2", auto)
      note res = res[simplified pa, simplified]
      from res obtain y4 s4 where oo2: "oneof [''/>'', ''>''] s3 = Inr (y4, s4)" by (cases "oneof [''/>'', ''>''] s3", auto)
      note res = res[unfolded oo2, simplified]
      from is_parser_parse_attributes and is_parser_oneof [of "[''/>'', ''>'']"]
        and is_cparser_exactly [of "''<''", simplified] and is_parser_parse_name
        and oo pn pa oo2
        have s_s4: "length s > length s4"
        by (auto dest!: is_parser_length is_cparser_length)
      show ?thesis
      proof (cases "y4 = ''/>''")
        case True
        from res True obtain y5
          where pns: "parse_nodes s4 = Inr (y5, r)"
          by (cases "parse_nodes s4") (auto simp: return_def)
        from ind2[OF s_s4 pns] s_s4 show "length r \<le> length s"  by simp
      next
        case False
        from res False obtain y5 s5 where pt: "parse_text s4 = Inr (y5, s5)" by (cases "parse_text s4", auto)
        note res = res[simplified False, simplified, unfolded pt, simplified,
          unfolded bind_def]
        from res obtain y6 s6 where pns: "parse_nodes s5 = Inr (y6, s6)" by (cases "parse_nodes s5", auto)
        note res = res[unfolded bind_def pns, simplified, unfolded bind_def]
        from res obtain y7 s7 where oo3: "exactly ''</'' s6 = Inr (y7, s7)" by (cases "exactly ''</'' s6", auto)
        note res = res[unfolded oo3, simplified, unfolded bind_def, 
          simplified, unfolded bind_def]
        from res obtain y8 s8 where oo4: "exactly y2 s7 = Inr (y8, s8)" by (cases "exactly y2 s7", auto)
        note res = res[unfolded oo4 bind_def, simplified]
        from res obtain y10 s10 where oo5: "exactly ''>'' s8 = Inr (y10,s10)"
          by (cases "exactly ''>'' s8", auto simp: bind_def)
        note res = res[unfolded oo5 bind_def, simplified]
        from res obtain y11 s11 where pns2: "parse_nodes s10 = Inr (y11, s11)" by (cases "parse_nodes s10", auto simp: bind_def)
        note res = res[unfolded bind_def pns2, simplified]
        from pt is_parser_parse_text s_s4
          have s_s5: "length s > length s5" by (auto dest: is_parser_length)
        note one = is_parser_oneof [unfolded is_parser_def, rule_format]
        note exact = is_parser_exactly [unfolded is_parser_def, rule_format]
        from ind2[OF s_s5 pns] s_s5 exact[OF oo3] exact[OF oo4]
          have s_s7: "length s > length s8" unfolding is_parser_def by force 
        with exact[OF oo5] have s_s10: "length s > length s10" by simp
        with ind2[OF s_s10 pns2] have s_s11: "length s > length s11" by simp
        thus "length r \<le> length s" using res by (auto simp: return_def)
      qed
    qed
  qed
qed simp

termination parse_nodes using parse_nodes_help by blast

lemma parse_nodes [intro]:
  "is_parser parse_nodes" 
  unfolding is_parser_def using parse_nodes_help by blast 

(* more efficient version of oneof [''/>'', ''>''] *)
fun oneof_closed :: "string parser" where
  "oneof_closed (x # xs) =
    (if x = CHR ''>'' then Error_Monad.return (''>'', trim xs)
    else if (x = CHR ''/'' \<and> (case xs of [] \<Rightarrow> False | Cons y ys \<Rightarrow> y = CHR ''>'')) then
      Error_Monad.return (''/>'', trim (tl xs))
    else err_expecting (''one of [/>, >]'') (x # xs))" |
  "oneof_closed xs = err_expecting (''one of [/>, >]'') xs"

lemma oneof_closed:
  "oneof [''/>'', ''>''] = oneof_closed" (is "?l = ?r")
proof (rule ext)
  fix xs
  have id: "''one of '' @ shows_list [''/>'',''>''] [] = ''one of [/>, >]''" by eval
  note d = oneof_def oneof_aux.simps id
  show "?l xs = ?r xs"
  proof (cases xs)
    case Nil
    show ?thesis unfolding Nil d by simp
  next
    case (Cons x xs) note oCons = this
    show ?thesis
    proof (cases "x = CHR ''>''")
      case True
      show ?thesis unfolding Cons d True by simp
    next
      case False note oFalse = this
      show ?thesis
      proof (cases "x = CHR ''/''")
        case False
        show ?thesis unfolding Cons d using False oFalse by simp
      next
        case True
        show ?thesis
        proof (cases xs)
          case Nil
          show ?thesis unfolding Cons Nil d by auto
        next
          case (Cons y ys)
          show ?thesis unfolding oCons Cons d by simp
        qed
      qed
    qed
  qed
qed

lemma If_removal:
  "(\<lambda> e x. if b e then f e x else g e x) = (\<lambda> e. if b e then f e else g e)"
  by (intro ext) auto

declare parse_nodes.simps [unfolded oneof_closed,
  unfolded If_removal [of "\<lambda> e. e = ''/>''"], code]

definition parse_node :: "xml parser" where
  "parse_node = do {
    exactly ''<'';
    n \<leftarrow> parse_name;
    atts \<leftarrow> parse_attributes;
    e \<leftarrow> oneof [''/>'', ''>''];
    if e = ''/>'' then return (XML n atts [] None)
    else do {
      t \<leftarrow> parse_text;
      cs \<leftarrow> parse_nodes;
      exactly ''</'';
      exactly n;
      exactly ''>'';
      return (XML n atts cs t)
    }}"

declare parse_node_def [unfolded oneof_closed, code]

function parse_header :: "string parser" where
  "parse_header ts =
    (if take 2 (trim ts) = ''<?'' then (do {
      h \<leftarrow> scan_upto ''?>'';
      hs \<leftarrow> parse_header;
      return (h @ hs)
    }) ts else (do {
      spaces;
      return []
    }) ts)"
  by (pat_completeness) auto

termination parse_header
proof
  fix ts y tsa
  assume "scan_upto ''?>'' ts = Inr (y, tsa)"
  with is_cparser_scan_upto have "length ts > length tsa"
    unfolding is_cparser_def by force
  thus "(tsa, ts) \<in> measure length" by simp
qed simp

consts scala :: bool

code_printing
  constant scala \<rightharpoonup>
    (Haskell) "False" and
    (Scala) "true" and
    (SML) "false" and
    (OCaml) "false" and
    (Eval) "false"

fun
  remove_comments_aux :: "bool \<Rightarrow> string \<Rightarrow> string"
where
  "remove_comments_aux False (c # cs) =
    (if (c = CHR ''<'' \<and> take 3 cs = ''!--'') then remove_comments_aux True (tl cs)
    else c # remove_comments_aux False cs)" |
  "remove_comments_aux True (c # cs) =
    (if (c = CHR ''-'' \<and> take 2 cs = ''->'') then remove_comments_aux False (drop 2 cs)
    else remove_comments_aux True cs)" |
  "remove_comments_aux _ [] = []"

fun remove_comments_aux_acc :: "string \<Rightarrow> bool \<Rightarrow> string \<Rightarrow> string" where
  "remove_comments_aux_acc a False (c # cs) =
    (if (c = CHR ''<'' \<and> take 3 cs = ''!--'') then remove_comments_aux_acc a True (tl cs)
    else remove_comments_aux_acc (c # a) False cs)" |
  "remove_comments_aux_acc a True (c # cs) =
    (if (c = CHR ''-'' \<and> take 2 cs = ''->'') then remove_comments_aux_acc a False (drop 2 cs)
    else remove_comments_aux_acc a True cs)" |
  "remove_comments_aux_acc a _ [] = a"

text {*A tail recursive variant for Scala to reduce stack size at the cost of double traversal.*}
definition remove_comments :: "string \<Rightarrow> string" where
  "remove_comments =
    (if scala then rev o (remove_comments_aux_acc [] False)
    else remove_comments_aux False)"

hide_const remove_comments_aux remove_comments_aux_acc

definition parse_doc :: "xmldoc parser" where
  "parse_doc = do {
    update_tokens remove_comments;
    h \<leftarrow> parse_header;
    xml \<leftarrow> parse_node;
    eoi;
    return (XMLDOC h xml)
  }"

definition doc_of_string :: "string \<Rightarrow> string + xmldoc" where
  "doc_of_string s = do {
    (doc, _) \<leftarrow> parse_doc s;
    Error_Monad.return doc
  }"

end

